/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "common.h"
#include "r5_pef.h"

static void ClearCommandArea(CommandArea* pCA)
{
    memset(pCA,0,sizeof(CommandArea) - 4);
    pCA->qChecksum = ~Crc32_ComputeBuf(
                   0,                       /* uint32_t inCrc32, */
                   pCA,                     /* const void *buf, */
                   sizeof(CommandArea) - 4  /* uint32_t bufLen */
                   );
}

void r5InitPEF(ProgEnvFuncs* pPEF)
{
    CommandArea* pCA;
    ResponseArea* pRA;
    LoopState* pLS;

    pLS = &pPEF->tLS;
    pLS->yFlashVerified = false;
    pLS->qMismatchCount = 0;

    pCA = &pPEF->pMA->tCommandArea;
    pRA = &pPEF->pMA->tResponseArea;

    ClearCommandArea(pCA);

    memset(pRA,0,sizeof(ResponseArea) - 8);
    pRA->qChecksum = ~Crc32_ComputeBuf(
                   0,                       /* uint32_t inCrc32, */
                   pRA,                     /* const void *buf, */
                   sizeof(ResponseArea) - 8  /* uint32_t bufLen */
                   );
    pRA->qFEEDC0DE = htole32(FEEDC0DE);

    pPEF->rDataBarrier(pPEF,CACHE_REGION_COMM_AREA);
    pPEF->rDataBarrier(pPEF,CACHE_REGION_RESP_AREA);
    pPEF->rDataBarrier(pPEF,CACHE_REGION_AUX_AREA);
    pPEF->rDataBarrier(pPEF,CACHE_REGION_PRIV_AREA);
}

/*
 * Return false if Step failed i.e. SIM_STOP is encountered
 */
bool r5StepPEF(ProgEnvFuncs* pPEF)
{
    CommandArea* pCA;
    ResponseArea* pRA;
    LoopState* pLS;
    bool yRet = true;

    pLS = &pPEF->tLS;
    pCA = &pPEF->pMA->tCommandArea;
    pRA = &pPEF->pMA->tResponseArea;

    pPEF->rDataBarrier(pPEF,CACHE_REGION_COMM_AREA);
    pLS->qCS = Crc32_ComputeBuf(
                   0,                       /* uint32_t inCrc32, */
                   pCA,                     /* const void *buf, */
                   sizeof(CommandArea) - 4  /* uint32_t bufLen */
                   );
    if(le32toh(pCA->qChecksum) != pLS->qCS) {
        if(0 == (pLS->qMismatchCount++ & 0xfffff)) {
            TRACE_PRINT("%08x %08x %08x %08x %08x %08x %08x %08x %08x ... %08x (Mismatch, expected %08x)\n\r",
                    pCA->qCommand,
                    pCA->pqArgs[0],
                    pCA->pqArgs[1],
                    pCA->pqArgs[2],
                    pCA->pqArgs[3],
                    pCA->pqArgs[4],
                    pCA->pqArgs[5],
                    pCA->pqArgs[6],
                    pCA->pqArgs[7],
                    le32toh(pCA->qChecksum),
                    pLS->qCS
                    );
        }
        return true;
    }
    TRACE_PRINT("%s %08x %08x %08x %08x %08x %08x %08x %08x ... %08x (Match)\n\r",
            GetCommandString(pCA->qCommand),
            pCA->pqArgs[0],
            pCA->pqArgs[1],
            pCA->pqArgs[2],
            pCA->pqArgs[3],
            pCA->pqArgs[4],
            pCA->pqArgs[5],
            pCA->pqArgs[6],
            pCA->pqArgs[7],
            pLS->qCS
            );
    pCA->qChecksum ^= ~0;
    pLS->qMismatchCount = 0;

    /* Each case must set ERR_NONE explicitly */
    pRA->qErrCode = htole32((uint32_t)ERR_INTERNAL); 
    pRA->qErrBSP = htole32((uint32_t)ERR_INTERNAL); 

    switch((COMMAND)le32toh(pCA->qCommand)) {
    default:
        pRA->qErrCode = htole32((uint32_t)ERR_COMMAND);
        pRA->qErrBSP = htole32(0);
        break;

    case COMMAND_VERIFY_FLASH:
        {
            pLS->tVerifyFlashArgsIn.qConnectionMode = le32toh(pCA->pqArgs[0]);
            pLS->tVerifyFlashArgsIn.qBusWidth = le32toh(pCA->pqArgs[1]);
            pLS->tVerifyFlashArgsIn.qQspiPeripheralInputSpeedHz = le32toh(pCA->pqArgs[2]);
            pLS->tVerifyFlashArgsIn.qPrescaler = le32toh(pCA->pqArgs[3]);
            pLS->tVerifyFlashArgsIn.qCacheCoherentDMA = le32toh(pCA->pqArgs[4]);
            pLS->tVerifyFlashArgsIn.qPEBSize = le32toh(pCA->pqArgs[5]);
            pLS->tVerifyFlashArgsIn.qNumPEBs = le32toh(pCA->pqArgs[6]);
            pLS->tVerifyFlashArgsIn.qSubsectorSize = le32toh(pCA->pqArgs[7]);
            if(pLS->tVerifyFlashArgsIn.qSubsectorSize != 0) {
                pLS->tVerifyFlashArgsIn.qNumSubsectors =
                    (pLS->tVerifyFlashArgsIn.qPEBSize * pLS->tVerifyFlashArgsIn.qNumPEBs)
                    /
                    pLS->tVerifyFlashArgsIn.qSubsectorSize;
            } else {
                pLS->tVerifyFlashArgsIn.qNumSubsectors = ~0;
            }

            if(pLS->yFlashVerified) {
                /*
                 * Called from a non-initial run of x86_app
                 * Simply check for errors
                 * QSPI is already running
                 */
                pRA->qErrBSP = htole32(0);
                if(0 == memcmp(&pLS->tVerifiedFlashArgs,&pLS->tVerifyFlashArgsIn,sizeof(VerifyFlashArgs))) {
                    pRA->qErrCode = htole32(ERR_NONE);
                } else {
                    pRA->qErrCode = htole32(ERR_BOARD_FLASH);
                }
            } else {
                if(0) {
                    ;
                } else if(0
                        || (0 == pLS->tVerifyFlashArgsIn.qPEBSize)
                        || (pLS->tVerifyFlashArgsIn.qPEBSize > MAX_PEB_SIZE_BYTES)
                    ) {
                    pRA->qErrCode = htole32(ERR_PEB_SIZE);
                    pRA->qErrBSP = htole32(0);
                } else if(0
                        || (0 == pLS->tVerifyFlashArgsIn.qSubsectorSize)
                        || (pLS->tVerifyFlashArgsIn.qSubsectorSize >= pLS->tVerifyFlashArgsIn.qPEBSize)
                        || (0 != (pLS->tVerifyFlashArgsIn.qPEBSize % pLS->tVerifyFlashArgsIn.qSubsectorSize))
                        || ((3*pLS->tVerifyFlashArgsIn.qSubsectorSize) > pLS->tVerifyFlashArgsIn.qPEBSize)
                    ) {
                    pRA->qErrCode = htole32(ERR_SUBSECTOR_SIZE);
                    pRA->qErrBSP = htole32(0);
                } else {
                    pRA->qErrCode = htole32(pPEF->rVerifyFlash(pPEF,&pRA->qErrBSP));
                    pRA->qErrBSP = htole32(pRA->qErrBSP);
                    if(htole32(ERR_NONE) == pRA->qErrCode) {
                        pLS->yFlashVerified = true;
                        memcpy(&pLS->tVerifiedFlashArgs,&pLS->tVerifyFlashArgsIn,sizeof(VerifyFlashArgs));
                    }
                }
            }
        }
        break;

    case COMMAND_VERIFY_IMAGE:
        if(pLS->yFlashVerified) {
            uint32_t qImageOffset;
            uint32_t qImageSize;
            uint32_t qImageChecksum;
            uint32_t qComputedChecksum;
            uint32_t qImagePEBs;
            uint32_t qPEBStart;

            qImageOffset = le32toh(pCA->pqArgs[0]);
            qImageSize = le32toh(pCA->pqArgs[1]);
            qImageChecksum = le32toh(pCA->pqArgs[2]);

            qImagePEBs = (qImageSize + pLS->tVerifiedFlashArgs.qPEBSize - 1) / pLS->tVerifiedFlashArgs.qPEBSize;
            qPEBStart = qImageOffset / pLS->tVerifiedFlashArgs.qPEBSize;
            if(qImageOffset % pLS->tVerifiedFlashArgs.qPEBSize) {
                pRA->qErrCode = htole32(ERR_IMAGE_ALIGNMENT);
            } else if(qPEBStart >= pLS->tVerifiedFlashArgs.qNumPEBs) {
                pRA->qErrCode = htole32(ERR_PEB_NUMBER);
            } else if(qImagePEBs > pLS->tVerifiedFlashArgs.qNumPEBs) {
                pRA->qErrCode = htole32(ERR_IMAGE_SIZE);
            } else if((qPEBStart + qImagePEBs) > pLS->tVerifiedFlashArgs.qNumPEBs) {
                pRA->qErrCode = htole32(ERR_IMAGE_SIZE);
            } else {
                uint32_t qPEB;
                uint32_t qBytesLeft;
                uint32_t qBytesToCompute;

                for(
                        qPEB = qPEBStart, qComputedChecksum = 0, qBytesLeft = qImageSize;
                        qBytesLeft > 0;
                        qPEB++
                    ) {
                    pRA->qErrCode = htole32(pPEF->rReadPEB(pPEF,qPEB,&pRA->qErrBSP,pPEF->pAA->pbData));
                    pRA->qErrBSP = htole32(pRA->qErrBSP);
                    if(pRA->qErrCode != le32toh(ERR_NONE)) {
                        break;
                    }

                    if(qBytesLeft > pLS->tVerifiedFlashArgs.qPEBSize) {
                        qBytesToCompute = pLS->tVerifiedFlashArgs.qPEBSize;
                        qBytesLeft -= pLS->tVerifiedFlashArgs.qPEBSize;
                    } else {
                        qBytesToCompute = qBytesLeft;
                        qBytesLeft = 0;
                    }
                    qComputedChecksum = Crc32_ComputeBuf(
                                   qComputedChecksum,   /* uint32_t inCrc32, */
                                   pPEF->pAA->pbData,   /* const void *buf, */
                                   qBytesToCompute      /* uint32_t bufLen */
                                   );
                }
                if((ERR_NONE == pRA->qErrCode) && (qComputedChecksum != qImageChecksum)) {
                    pRA->qErrCode = htole32(ERR_CHECKSUM);
                    pRA->qErrBSP = htole32(qComputedChecksum);
                }
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_GET_MULTIBOOT_BYTES:
        if(pLS->yFlashVerified) {
            uint32_t qMultibootBytes;

            pRA->qErrCode = htole32(pPEF->rGetMultibootBytes(pPEF,&pRA->qErrBSP,&qMultibootBytes));
            pRA->qErrBSP = htole32(pRA->qErrBSP);
            if(1
                && (le32toh(ERR_NONE) == pRA->qErrCode)
                && (
                    (qMultibootBytes > (pLS->tVerifiedFlashArgs.qPEBSize * pLS->tVerifiedFlashArgs.qNumPEBs))
                    ||
                    (0 != (qMultibootBytes % pLS->tVerifiedFlashArgs.qPEBSize))
                    )
              ) {
                pRA->qErrCode = htole32(ERR_MULTIBOOT_BYTES);
            }
            pRA->pqData[0] = htole32(qMultibootBytes);
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_ERASE_PEBS:
        if(pLS->yFlashVerified) {
            uint32_t qPEBStart;
            uint32_t qPEBEnd;

            qPEBStart = le32toh(pCA->pqArgs[0]);
            qPEBEnd = le32toh(pCA->pqArgs[1]);

            if((qPEBEnd < qPEBStart) || (qPEBStart >= pLS->tVerifiedFlashArgs.qNumPEBs) || (qPEBEnd >= pLS->tVerifiedFlashArgs.qNumPEBs)) {
                pRA->qErrCode = htole32(ERR_PEB_NUMBER);
                pRA->qErrBSP = htole32(0);
            } else {
                pRA->qErrCode = htole32(pPEF->rErasePEBs(pPEF,qPEBStart,qPEBEnd,&pRA->qErrBSP));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_READ_PEB:
        if(pLS->yFlashVerified) {
            uint32_t qPEBNumber;

            qPEBNumber = le32toh(pCA->pqArgs[0]);
            if(qPEBNumber >= pLS->tVerifiedFlashArgs.qNumPEBs) {
                pRA->qErrCode = htole32(ERR_PEB_NUMBER);
                pRA->qErrBSP = htole32(0);
            } else {
                pPEF->rDataBarrier(pPEF,CACHE_REGION_AUX_AREA);
                pRA->qErrCode = htole32(pPEF->rReadPEB(pPEF,qPEBNumber,&pRA->qErrBSP,pPEF->pAA->pbData));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_WRITE_PEB:
        if(pLS->yFlashVerified) {
            uint32_t qPEBNumber;

            qPEBNumber = le32toh(pCA->pqArgs[0]);
            if(qPEBNumber >= pLS->tVerifiedFlashArgs.qNumPEBs) {
                pRA->qErrCode = htole32(ERR_PEB_NUMBER);
                pRA->qErrBSP = htole32(0);
            } else {
                pPEF->rDataBarrier(pPEF,CACHE_REGION_AUX_AREA);
                pRA->qErrCode = htole32(pPEF->rWritePEB(pPEF,qPEBNumber,pPEF->pAA->pbData,&pRA->qErrBSP));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_ERASE_SUBSECTOR:
        if(pLS->yFlashVerified) {
            uint32_t qSubsectorNumber;
            uint32_t qPEB;
            uint32_t qByteOffset;

            qSubsectorNumber = le32toh(pCA->pqArgs[0]);
            if(qSubsectorNumber >= pLS->tVerifiedFlashArgs.qNumSubsectors) {
                pRA->qErrCode = htole32(ERR_SUBSECTOR_NUMBER);
                pRA->qErrBSP = htole32(0);
            } else {
                qByteOffset = qSubsectorNumber * pLS->tVerifiedFlashArgs.qSubsectorSize;
                qPEB = qByteOffset / pLS->tVerifiedFlashArgs.qPEBSize;
                qByteOffset %= pLS->tVerifiedFlashArgs.qPEBSize;

                pPEF->rDataBarrier(pPEF,CACHE_REGION_PRIV_AREA);
                pRA->qErrCode = htole32(pPEF->rReadPEB(pPEF,qPEB,&pRA->qErrBSP,pPEF->pPA->pbData));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
                if(le32toh(ERR_NONE) != pRA->qErrCode) {
                    break;
                }

                pRA->qErrCode = htole32(pPEF->rErasePEBs(pPEF,qPEB,qPEB,&pRA->qErrBSP));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
                if(le32toh(ERR_NONE) != pRA->qErrCode) {
                    break;
                }

                memset(&pPEF->pPA->pbData[qByteOffset],0xff,pLS->tVerifiedFlashArgs.qSubsectorSize);

                pPEF->rDataBarrier(pPEF,CACHE_REGION_PRIV_AREA);
                pRA->qErrCode = htole32(pPEF->rWritePEB(pPEF,qPEB,pPEF->pPA->pbData,&pRA->qErrBSP));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_READ_SUBSECTOR:
        if(pLS->yFlashVerified) {
            uint32_t qSubsectorNumber;
            uint32_t qPEB;
            uint32_t qByteOffset;

            qSubsectorNumber = le32toh(pCA->pqArgs[0]);
            if(qSubsectorNumber >= pLS->tVerifiedFlashArgs.qNumSubsectors) {
                pRA->qErrCode = htole32(ERR_SUBSECTOR_NUMBER);
                pRA->qErrBSP = htole32(0);
            } else {
                qByteOffset = qSubsectorNumber * pLS->tVerifiedFlashArgs.qSubsectorSize;
                qPEB = qByteOffset / pLS->tVerifiedFlashArgs.qPEBSize;
                qByteOffset %= pLS->tVerifiedFlashArgs.qPEBSize;

                pPEF->rDataBarrier(pPEF,CACHE_REGION_PRIV_AREA);
                pRA->qErrCode = htole32(pPEF->rReadPEB(pPEF,qPEB,&pRA->qErrBSP,pPEF->pPA->pbData));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
                if(le32toh(ERR_NONE) != pRA->qErrCode) {
                    break;
                }
                
                pPEF->rDataBarrier(pPEF,CACHE_REGION_AUX_AREA);
                memcpy(pPEF->pAA->pbData,&pPEF->pPA->pbData[qByteOffset],pLS->tVerifiedFlashArgs.qSubsectorSize);
                pPEF->rDataBarrier(pPEF,CACHE_REGION_AUX_AREA);
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_WRITE_SUBSECTOR:
        if(pLS->yFlashVerified) {
            uint32_t qSubsectorNumber;
            uint32_t qPEB;
            uint32_t qByteOffset;

            qSubsectorNumber = le32toh(pCA->pqArgs[0]);
            if(qSubsectorNumber >= pLS->tVerifiedFlashArgs.qNumSubsectors) {
                pRA->qErrCode = htole32(ERR_SUBSECTOR_NUMBER);
                pRA->qErrBSP = htole32(0);
            } else {
                pPEF->rDataBarrier(pPEF,CACHE_REGION_AUX_AREA);

                qByteOffset = qSubsectorNumber * pLS->tVerifiedFlashArgs.qSubsectorSize;
                qPEB = qByteOffset / pLS->tVerifiedFlashArgs.qPEBSize;
                qByteOffset %= pLS->tVerifiedFlashArgs.qPEBSize;

                pPEF->rDataBarrier(pPEF,CACHE_REGION_PRIV_AREA);
                pRA->qErrCode = htole32(pPEF->rReadPEB(pPEF,qPEB,&pRA->qErrBSP,pPEF->pPA->pbData));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
                if(le32toh(ERR_NONE) != pRA->qErrCode) {
                    break;
                }

                pRA->qErrCode = htole32(pPEF->rErasePEBs(pPEF,qPEB,qPEB,&pRA->qErrBSP));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
                if(le32toh(ERR_NONE) != pRA->qErrCode) {
                    break;
                }

                memcpy(&pPEF->pPA->pbData[qByteOffset],pPEF->pAA->pbData,pLS->tVerifiedFlashArgs.qSubsectorSize);
                pPEF->rDataBarrier(pPEF,CACHE_REGION_PRIV_AREA);

                pRA->qErrCode = htole32(pPEF->rWritePEB(pPEF,qPEB,pPEF->pPA->pbData,&pRA->qErrBSP));
                pRA->qErrBSP = htole32(pRA->qErrBSP);
            }
        } else {
            pRA->qErrCode = htole32(ERR_FLASH_UNVERIFIED);
            pRA->qErrBSP = htole32(0);
        }
        break;

    case COMMAND_SIM_STOP:
        pRA->qErrCode = htole32(ERR_NONE);
        pRA->qErrBSP = htole32(0);
        yRet = false;
        break;
    }
    ClearCommandArea(pCA);
    pRA->qChecksum = Crc32_ComputeBuf(
                   0,                       /* uint32_t inCrc32, */
                   pRA,                     /* const void *buf, */
                   sizeof(ResponseArea) - 8 /* uint32_t bufLen */
                   );
    pPEF->rDataBarrier(pPEF,CACHE_REGION_RESP_AREA);

    return yRet;
}
