/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <x86_app.h>

typedef bool (*CheckFunc)(ListHead* pListHead,ListElem* pLE);

/*
 * --------------------------------------------------------------
 * Parse csv file into a list of user-defined structures
 * --------------------------------------------------------------
 */
static bool ParseCsv(
        Sys* pSys,
        ListHead* pListHead,
        int iExpectedTokens,
        const char* pcSep,
        uint32_t qAllocBytes,
        uint32_t qCsvRowOffset,
            /*
             * Validate row before inserting in list
             */
        CheckFunc rCheckFunc
        )
{
    FILE* pFile;
    uint8_t* pbBuf;
    CsvRow* pCR;
    int iLine;
    int iErr;

    iLine = 0;
    if(iExpectedTokens > CSV_MAX_TOKENS) {
        Die(pSys,errno,"%s:%d: iExpectedTokens=%d > %d",
                pListHead->pcFile,
                iLine,
                iExpectedTokens,
                CSV_MAX_TOKENS
                );
        /* Not reached */
        return false;
    }
    pFile = fopen(pListHead->pcFile,"r");
    if(!pFile) {
        Die(pSys,errno,"%s",pListHead->pcFile);
        /* Not reached */
        return false;
    }
    while(!feof(pFile)) {
        iErr = posix_memalign((void**)&pbBuf,16,qAllocBytes);
        if(0 != iErr) {
            FreeList(pListHead);
            fclose(pFile);
            Die(pSys,iErr,"%d",qAllocBytes);
            /* Not reached */
            return false;
        }
        memset(pbBuf,0,qAllocBytes);

        pCR = (CsvRow*)(pbBuf + qCsvRowOffset);
        pCR->iLine = ++iLine;
        pCR->pcLine[CSV_LINE_BYTES - 2] = 0;
        if(!fgets(pCR->pcLine,CSV_LINE_BYTES,pFile)) {
            break;
        } else if(1
                && (0 != pCR->pcLine[CSV_LINE_BYTES - 2])
                && ('\n' != pCR->pcLine[CSV_LINE_BYTES - 2])
            ) {
            FreeList(pListHead);
            fclose(pFile);
            Die(pSys,errno,"%s:%d: Line too long",
                    pListHead->pcFile,
                    iLine
                    );
            /* Not reached */
            return false;
        }
        pCR->iNumTokens = Tokenize(
                pSys,
                pListHead->pcFile,
                pCR->iLine,
                pCR->pcLine,
                pcSep,
                iExpectedTokens,
                pCR->ppcTokens
                );
        if(!pCR->iNumTokens) {
            free(pbBuf);
            continue;
        }
        if(pCR->iNumTokens != iExpectedTokens) {
            FreeList(pListHead);
            fclose(pFile);
            Die(pSys,0,"%s:%d: Expected %d, got %d tokens",
                    pListHead->pcFile,
                    iLine,
                    iExpectedTokens,
                    pCR->iNumTokens
                    );
            /* Not reached */
            return false;
        }
        pCR->ppcTokens[iExpectedTokens] = 0;

        if(rCheckFunc && !rCheckFunc(pListHead,(ListElem*)pbBuf)) {
            FreeList(pListHead);
            fclose(pFile);
            Die(pSys,0,"%s:%d: Invalid line",pListHead->pcFile,iLine);
            /* Not reached */
            return false;
        }

        InsertListElem(pListHead,(ListElem*)pbBuf);
    }
    fclose(pFile);
    return true;
}

/*
 * --------------------------------------------------------------
 * ParseImagesCsv and helper functions
 * --------------------------------------------------------------
 */
bool CheckImagesCsvRow(ListHead* pListHead,ListElem* pLE)
{
    ImageInfo* pII = (ImageInfo*)pLE;
    struct stat tStat;
    int iFile;
    int errnosave;

    if(pII->tCR.ppcTokens[0][0] == '/') {
        Die(pListHead->pSys,0,"%s:%d: Not a relative pathname: '%s'",
                pListHead->pcFile,
                pII->tCR.iLine,
                pII->tCR.ppcTokens[0]
                );
        /* Not reached */
        return false;
    }
    if((pListHead->pSys->qImagesPathLen + 1 + strlen(pII->tCR.ppcTokens[0] + 1)) > II_FULLNAME_BYTES) {
        Die(pListHead->pSys,0,"%s:%d: Fullpath too long: '%s/%s'",
                pListHead->pcFile,
                pII->tCR.iLine,
                pListHead->pSys->pcImagesPath,
                pII->tCR.ppcTokens[0]
                );
        /* Not reached */
        return false;
    }
    sprintf(pII->pcFullName,"%s/%s",pListHead->pSys->pcImagesPath,pII->tCR.ppcTokens[0]);

    if(-1 == stat(pII->pcFullName,&tStat)) {
        Die(0,errno,"Image file '%s'",pII->pcFullName);
        /* Not reached */
        return false;
    } else if(!S_ISREG(tStat.st_mode)) {
        Die(0,errno,"Image file '%s' is not a regular file",pII->pcFullName);
        /* Not reached */
        return false;
    }
    iFile = open(pII->pcFullName,O_RDONLY);
    if(iFile < 0) {
        Die(0,errno,"open('%s')",pII->pcFullName);
        /* Not reached */
        return false;
    }

    pII->qSize = (uint32_t)tStat.st_size;
    pII->pxMap = mmap(
                0,                          /* addr */
                pII->qSize,                 /* length */
                PROT_READ,                  /* prot */
                MAP_PRIVATE,                /* flags */
                iFile,                      /* fd */
                0                           /* offset */
                );
    errnosave = errno;

    if(0 != close(iFile)) {
        Die(pListHead->pSys,errno,"close('%s')",pII->pcFullName);
        /* Not reached */
        return false;
    }

    if(MAP_FAILED == pII->pxMap) {
        Die(pListHead->pSys,errnosave,"mmap('%s')",pII->pcFullName);
        /* Not reached */
        return false;
    }

    pII->qChecksum = Crc32_ComputeBuf(
                            0,              /* uint32_t inCrc32, */
                            pII->pxMap,          /* const void *buf, */
                            pII->qSize      /* uint32_t bufLen */
                        );

    if(0) {
        ;
    } else if(0 == strcasecmp(pII->tCR.ppcTokens[1],"prom")) {
        pII->eType = IMAGE_TYPE_PROM;
    } else if(0 == strcasecmp(pII->tCR.ppcTokens[1],"stage1")) {
        pII->eType = IMAGE_TYPE_STAGE1;
    } else if(0 == strcasecmp(pII->tCR.ppcTokens[1],"stage2")) {
        pII->eType = IMAGE_TYPE_STAGE2;
    } else {
        Die(pListHead->pSys,0,"%s:%d: Invalid stage type: '%s', must be one of 'prom', 'stage1', or 'stage2'",
                pListHead->pcFile,
                pII->tCR.iLine,
                pII->tCR.ppcTokens[1]
                );
        return false;
    }

    if(strlen(pII->tCR.ppcTokens[2]) != 128) {
        Die(pListHead->pSys,0,"%s:%d: Length must be 128 bytes: '%s'",
                pListHead->pcFile,
                pII->tCR.iLine,
                pII->tCR.ppcTokens[2]
                );
        /* Not reached */
        return false;
    }
    return true;
}

bool ParseImagesCsv(Sys* pSys,ListHead* pListHead)
{
    return ParseCsv(pSys,pListHead,3,",\n",sizeof(ImageInfo),offsetof(ImageInfo,tCR),CheckImagesCsvRow);
}

/*
 * --------------------------------------------------------------
 * ParseBoardsCsv and helper functions
 * --------------------------------------------------------------
 */
bool CheckBoardsCsvRow(ListHead* pListHead,ListElem* pLE)
{
    BoardInfo* pBI = (BoardInfo*)pLE;
    const uint32_t FREQ_37_DOT_5_MHZ = 37500000;

    if(!CheckBoardTypeSyntax(pListHead->pSys,pListHead->pcFile,pBI->tCR.iLine,pBI->tCR.ppcTokens[0])) {
        return false;
    }

    if(0) {
        ;
    } else if(0 == strcasecmp(pBI->tCR.ppcTokens[3],"lower")) {
        pBI->qConnectionMode = CONNECTION_MODE_LOWER;
    } else if(0 == strcasecmp(pBI->tCR.ppcTokens[3],"dualParallel")) {
        pBI->qConnectionMode = CONNECTION_MODE_DUAL_PARALLEL;
    } else {
        Die(pListHead->pSys,0,
                "%s:%d: '%s': ConnectionMode must be one of 'lower', 'dualParallel'",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->tCR.ppcTokens[3]
                );
        return false;
    }

    if(0
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "R5CommBar",
                pBI->tCR.ppcTokens[1],
                10,
                &pBI->qR5CommBar)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "TCMOffset",
                pBI->tCR.ppcTokens[2],
                10,
                &pBI->qTCMOffset)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "BusWidth",
                pBI->tCR.ppcTokens[4],
                10,
                &pBI->qBusWidth)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "QspiPeripheralInputSpeedHz",
                pBI->tCR.ppcTokens[5],
                10,
                &pBI->qQspiPeripheralInputSpeedHz)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "QspiBusMaxSpeedHz",
                pBI->tCR.ppcTokens[6],
                10,
                &pBI->qQspiBusMaxSpeedHz)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "CacheCoherentDMA",
                pBI->tCR.ppcTokens[7],
                10,
                &pBI->qCacheCoherentDMA)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "PEBSize",
                pBI->tCR.ppcTokens[8],
                10,
                &pBI->qPEBSize)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "NumPebs",
                pBI->tCR.ppcTokens[9],
                10,
                &pBI->qNumPEBs)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "SubsectorSize",
                pBI->tCR.ppcTokens[10],
                10,
                &pBI->qSubsectorSize)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "PEBOffsetB",
                pBI->tCR.ppcTokens[11],
                10,
                &pBI->qPEBOffsetB)
        || !GetBaseNumber(
                pListHead->pSys,
                pListHead->pcFile,
                pBI->tCR.iLine,
                "PEBOffsetG",
                pBI->tCR.ppcTokens[12],
                10,
                &pBI->qPEBOffsetG)
      ) {
        return false;
    }

    if(pBI->qR5CommBar > 5) {
        Die(pListHead->pSys,0,"%s:%d: '%s': R5CommBar must be less than 6",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->tCR.ppcTokens[1]
                );
        return false;
    }

    switch(pBI->qBusWidth) {
    case 1: /* FallThrough */
    case 2: /* FallThrough */
    case 4:
        break;

    default:
        Die(pListHead->pSys,0,"%s:%d: '%s': BusWidth must be 1, 2 or 4",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->tCR.ppcTokens[4]
                );
        return false;
    }

    {
        uint32_t qSpeed;

        if(0 == pBI->qQspiPeripheralInputSpeedHz) {
            Die(pListHead->pSys,0,"%s:%d: '%s': QspiPeripheralInputSpeedHz must be positive",
                    pListHead->pcFile,
                    pBI->tCR.iLine,
                    pBI->tCR.ppcTokens[5]
                    );
            return false;
        }

        if(0 == pBI->qQspiBusMaxSpeedHz) {
            Die(pListHead->pSys,0,"%s:%d: '%s': QspiBusMaxSpeedHz must be positive",
                    pListHead->pcFile,
                    pBI->tCR.iLine,
                    pBI->tCR.ppcTokens[6]
                    );
            return false;
        }

        for(pBI->qPrescaler = 2; pBI->qPrescaler < 512; pBI->qPrescaler *= 2) {
            const char* pcReason;
            const char* pcSpeedType;

            qSpeed = pBI->qQspiPeripheralInputSpeedHz / pBI->qPrescaler;
            if(0 == qSpeed) {
                Die(pListHead->pSys,0,
                    "%s:%d: %s: Zero QSPI bus speed(!) for input clock %u and prescaler %u",
                    pListHead->pcFile,
                    pBI->tCR.iLine,
                    pBI->tCR.ppcTokens[0],
                    pBI->qQspiPeripheralInputSpeedHz,
                    pBI->qPrescaler
                    );
                return false;
            } else if(qSpeed > FREQ_37_DOT_5_MHZ) {
                pcSpeedType = "high";
            } else {
                pcSpeedType = "low";
            }
            if(qSpeed <= pBI->qQspiBusMaxSpeedHz) {
                if(2 == pBI->qPrescaler) {
                    pcReason = "GOOD: in range, valid prescaler";
                } else if(qSpeed <= FREQ_37_DOT_5_MHZ) {
                    pcReason = "GOOD: in range";
                } else {
                    pcReason = "BAD: in range, but prescaler is not 2";
                }
            } else {
                pcReason = "BAD: Too fast";
            }
            Verbose(pListHead->pSys,
                        "%s: Input: %u, Scaler: %u, QSPI bus speed: %u: %s (%s speed)",
                        pBI->tCR.ppcTokens[0],
                        pBI->qQspiPeripheralInputSpeedHz,
                        pBI->qPrescaler,
                        qSpeed,
                        pcReason,
                        pcSpeedType
                        );
            if('G' == *pcReason) {
                break;
            }
        }
        if(pBI->qPrescaler >= 512) {
            Die(pListHead->pSys,0,
                "%s:%d: %s: Desired speed %u is smaller than lowest achievable speed %u as input clock is %u",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->tCR.ppcTokens[0],
                pBI->qQspiBusMaxSpeedHz,
                qSpeed,
                pBI->qQspiPeripheralInputSpeedHz
                );
            return false;
        }
        if(qSpeed < pBI->qQspiBusMaxSpeedHz) {
            Warn(pListHead->pSys,0,
                "%s:%d: %s: Reducing QSPI bus speed from %u to %u as input clock is %u",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->tCR.ppcTokens[0],
                pBI->qQspiBusMaxSpeedHz,
                qSpeed,
                pBI->qQspiPeripheralInputSpeedHz
                );
        }
    }

    if(1
            && (0 != pBI->qCacheCoherentDMA)
            && (1 != pBI->qCacheCoherentDMA)
        ) {
        Die(pListHead->pSys,0,"%s:%d: '%s': CacheCoherentDMA must be 0 or 1",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->tCR.ppcTokens[7]
                );
        return false;
    }

    if(0 == pBI->qPEBSize) {
        Die(pListHead->pSys,0,"%s:%d PEBSize is zero",
                pListHead->pcFile,
                pBI->tCR.iLine
                );
        return false;
    }
    if(0 == pBI->qNumPEBs) {
        Die(pListHead->pSys,0,"%s:%d NumPEBs is zero",
                pListHead->pcFile,
                pBI->tCR.iLine
                );
        return false;
    }
    if(pBI->qSubsectorSize < sizeof(ReservedBlockEntry)) {
        Die(pListHead->pSys,0,"%s:%d SubsectorSize %u must be at least %u (ReservedBlockEntry size)",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->qSubsectorSize,
                (uint32_t)sizeof(ReservedBlockEntry)
                );
        return false;
    }
    if((3*pBI->qSubsectorSize) > pBI->qPEBSize) {
        Die(pListHead->pSys,0,"%s:%d SubsectorSize %u too big for PEBSize %u: at least 3 must fit in a PEB",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->qSubsectorSize,
                pBI->qPEBSize
                );
        return false;
    }
    if(((pBI->qPEBSize / pBI->qSubsectorSize) * pBI->qSubsectorSize) != pBI->qPEBSize) {
        Die(pListHead->pSys,0,"%s:%d PEB size %u not a multiple of SubsectorSize %u",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->qPEBSize,
                pBI->qSubsectorSize
                );
        return false;
    }
    if((2*pBI->qPEBOffsetB) > pBI->qPEBOffsetG) {
        Die(pListHead->pSys,0,"%s:%d PEBOffsetG=%u must be at least 2*PEBOffsetB=%u",
                pListHead->pcFile,
                pBI->tCR.iLine,
                pBI->qPEBOffsetG,
                pBI->qPEBOffsetB
                );
        return false;
    }
    return true;
}

bool ParseBoardsCsv(Sys* pSys,ListHead* pListHead)
{
    return ParseCsv(pSys,pListHead,13,", \t\n",sizeof(BoardInfo),offsetof(BoardInfo,tCR),CheckBoardsCsvRow);
}
