/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <x86_app.h>

static bool FindImage(ListHead* pListHead,ListElem* pListElem,void* pxArg)
{
    ImageInfo* pII = (ImageInfo*)pListElem;
    uint32_t* pqChecksum = (uint32_t*)pxArg;

    return pII->qChecksum == *pqChecksum;
}

void DecodeDevInfo(Sys* pSys)
{
    int iRBE;
    struct BoardInfo* pBI = pSys->pBI;
    uint32_t qMultibootBytes;
    uint32_t qSubsector;
    ReservedBlockEntry* pRBE;
    ImageInfo* pII;
    uint32_t qErrCode;
    uint32_t qErrBSP;

    if(!pBI) {
        Warn(pSys,0,"No device selected");
        return;
    }
    qErrCode = GetMultibootBytes(pSys,&qErrBSP,&qMultibootBytes);
    if(ERR_NONE != qErrCode) {
        Die(pSys,0,"GetMultibootBytes() -> %u(%s) (ErrBSP=%u)",qErrCode,GetErrString(qErrCode),qErrBSP);
        /* Not reached */
        return;
    }

    qSubsector = (pBI->qPEBSize/pBI->qSubsectorSize) * (pBI->qNumPEBs - 1);

    pSys->qMultibootChar = (uint32_t)'x';
    for(iRBE = 0; iRBE < 3; iRBE++) {
        pRBE = pSys->ppRBE[iRBE];
        pRBE->qErrCode = ReadSubsector(pSys, qSubsector + iRBE, &pRBE->qErrBSP, (uint8_t*)pRBE);
        if(ERR_NONE != pRBE->qErrCode) {
            Warn(pSys,0,"ReadSubsector(%u) -> %u(%s) (ErrBSP = %d)",
                    qSubsector + iRBE,
                    pRBE->qErrCode,
                    GetErrString(pRBE->qErrCode),
                    pRBE->qErrBSP
                    );
            continue;
        }

        pRBE->qVersion = le32toh(pRBE->qVersion);
        pRBE->qImageOffset = le32toh(pRBE->qImageOffset);
        pRBE->qImageSize = le32toh(pRBE->qImageSize);
        pRBE->qImageChecksum = le32toh(pRBE->qImageChecksum);
        pRBE->qEntryChecksum = le32toh(pRBE->qEntryChecksum);

        if(pRBE->qEntryChecksum != Crc32_ComputeBuf( 
                                0,                            /* uint32_t inCrc32, */
                                (void*)pRBE,                  /* const void *buf, */
                                sizeof(ReservedBlockEntry) - 4    /* uint32_t bufLen */
                                )) {
            Warn(pSys,0,"Reserved Block Entry #%d -> bad checksum, ignoring",iRBE);
            pRBE->qErrCode = (uint32_t)ERR_CHECKSUM;
            continue;
        }

        if(RBE_VERSION != pRBE->qVersion) {
            Warn(pSys,0,"Reserved Block Entry #%d -> version %u, expected %u, ignoring",
                    iRBE,pRBE->qVersion,RBE_VERSION);
            pRBE->qErrCode = (uint32_t)ERR_CHECKSUM;
            continue;
        }

        pRBE->qErrCode = VerifyImage(
                pSys,
                pRBE->qImageOffset,
                pRBE->qImageSize,
                pRBE->qImageChecksum,
                &pRBE->qErrBSP
                );
        if(ERR_NONE == pRBE->qErrCode) {
            if(qMultibootBytes == pRBE->qImageOffset) {
                switch(iRBE) {
                default:    /* FallThrough */
                case 0:
                    pSys->qMultibootChar = (uint32_t)'a';
                    break;
                case 1:
                    pSys->qMultibootChar = (uint32_t)'b';
                    break;
                case 2:
                    pSys->qMultibootChar = (uint32_t)'g';
                    break;
                }
            }
        } else {
            Warn(pSys,0,"Reserved Block Entry #%d -> %s (ErrBSP=%u)",iRBE,GetErrString(qErrCode),pRBE->qErrBSP);
        }
    }

    for(iRBE = 0; iRBE < 3; iRBE++) {
        pRBE = pSys->ppRBE[iRBE];
        pII = (ImageInfo*)CheckElems(&pSys->tImagesCsvListHead,FindImage,&pRBE->qImageChecksum);

        Info(pSys,"%c: Image@0x%08x, size 0x%08x, ExpectedChecksum 0x%08x: %s(ErrBSP=%u): %s%s",
                toupper((0 == iRBE)?'a':((1 == iRBE)?'b':'g')),
                pRBE->qImageOffset,
                pRBE->qImageSize,
                pRBE->qImageChecksum,
                GetErrString(pRBE->qErrCode),
                pRBE->qErrBSP,
                pII?(char*)(pII->tCR.ppcTokens[0]):"Not in images csv file ",
                pII?((pRBE->qErrBSP || pRBE->qErrCode)?"(Expected but not found)":""):pSys->pcImagesCsv
                );
    }

    if((uint32_t)'x' == pSys->qMultibootChar) {
        Warn(pSys,0,"Error: qMultiboot=0x%08x does not match any image",qMultibootBytes);
    } else {
        Info(pSys,"Multiboot Image: %c (0x%08x)",toupper(pSys->qMultibootChar),qMultibootBytes);
    }
}
